@@ -3,34 +3,26 @@
 # A connection pool allowing multi-threaded access to a pool of connections.
 # This is the default connection pool used by Sequel.
 class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
-  # Whether or not a ConditionVariable should be used to wait for connections.
-  # True except on ruby 1.8, where ConditionVariable#wait does not support a
-  # timeout.
-  unless defined?(USE_WAITER)
-    USE_WAITER = RUBY_VERSION >= '1.9'
-  end
+  USE_WAITER = true
+  Sequel::Deprecation.deprecate_constant(self, :USE_WAITER)
 
   # The maximum number of connections this pool will create (per shard/server
   # if sharding).
   attr_reader :max_size
   
   # An array of connections that are available for use by the pool.
+  # The calling code should already have the mutex before calling this.
   attr_reader :available_connections
   
-  # A hash with thread keys and connection values for currently allocated
-  # connections.
+  # A hash with thread keys and connection values for currently allocated connections.
+  # The calling code should already have the mutex before calling this.
   attr_reader :allocated
 
   # The following additional options are respected:
-  # * :connection_handling - Set how to handle available connections.  By default,
-  #   uses a a queue for fairness.  Can be set to :stack to use a stack, which may
-  #   offer better performance.
-  # * :max_connections - The maximum number of connections the connection pool
-  #   will open (default 4)
-  # * :pool_sleep_time - The amount of time to sleep before attempting to acquire
-  #   a connection again, only used on ruby 1.8. (default 0.001)
-  # * :pool_timeout - The amount of seconds to wait to acquire a connection
-  #   before raising a PoolTimeoutError (default 5)
+  # :max_connections :: The maximum number of connections the connection pool
+  #                     will open (default 4)
+  # :pool_timeout :: The amount of seconds to wait to acquire a connection
+  #                  before raising a PoolTimeoutError (default 5)
   def initialize(db, opts = OPTS)
     super
     @max_size = Integer(opts[:max_connections] || 4)
@@ -40,14 +32,7 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
     @available_connections = []
     @allocated = {}
     @timeout = Float(opts[:pool_timeout] || 5)
-
-    if USE_WAITER
-      @waiter = ConditionVariable.new
-    else
-      # :nocov:
-      @sleep_time = Float(opts[:pool_sleep_time] || 0.001)
-      # :nocov:
-    end
+    @waiter = ConditionVariable.new
   end
   
   # Yield all of the available connections, and the one currently allocated to
@@ -70,15 +55,18 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
   # being used.  If you want to be able to disconnect connections that are
   # currently in use, use the ShardedThreadedConnectionPool, which can do that.
   # This connection pool does not, for performance reasons. To use the sharded pool,
-  # pass the <tt>:servers=>{}</tt> option when connecting to the database.
+  # pass the <tt>servers: {}</tt> option when connecting to the database.
   # 
   # Once a connection is requested using #hold, the connection pool
   # creates new connections to the database.
   def disconnect(opts=OPTS)
+    conns = nil
     sync do
-      @available_connections.each{|conn| db.disconnect_connection(conn)}
+      conns = @available_connections.dup
       @available_connections.clear
+      @waiter.signal
     end
+    conns.each{|conn| disconnect_connection(conn)}
   end
   
   # Chooses the first available connection, or if none are
@@ -93,8 +81,7 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
   # If no connection is immediately available and the pool is already using the maximum
   # number of connections, Pool#hold will block until a connection
   # is available or the timeout expires.  If the timeout expires before a
-  # connection can be acquired, a Sequel::PoolTimeout is 
-  # raised.
+  # connection can be acquired, a Sequel::PoolTimeout is raised.
   def hold(server=nil)
     t = Thread.current
     if conn = owned_connection(t)
@@ -103,14 +90,24 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
     begin
       conn = acquire(t)
       yield conn
-    rescue Sequel::DatabaseDisconnectError
-      oconn = conn
-      conn = nil
-      db.disconnect_connection(oconn) if oconn
-      @allocated.delete(t)
+    rescue Sequel::DatabaseDisconnectError, *@error_classes => e
+      if disconnect_error?(e)
+        oconn = conn
+        conn = nil
+        disconnect_connection(oconn) if oconn
+        sync do 
+          @allocated.delete(t)
+          @waiter.signal
+        end
+      end
       raise
     ensure
-      sync{release(t)} if conn
+      if conn
+        sync{release(t)}
+        if @connection_handling == :disconnect
+          disconnect_connection(conn)
+        end
+      end
     end
   end
 
@@ -119,111 +116,115 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
   end
   
   # The total number of connections opened, either available or allocated.
-  # This may not be completely accurate as it isn't protected by the mutex.
+  # The calling code should not have the mutex before calling this.
   def size
-    @allocated.length + @available_connections.length
+    @mutex.synchronize{_size}
   end
   
   private
 
+  # The total number of connections opened, either available or allocated.
+  # The calling code should already have the mutex before calling this.
+  def _size
+    @allocated.length + @available_connections.length
+  end
+
   # Assigns a connection to the supplied thread, if one
-  # is available. The calling code should already have the mutex when
+  # is available. The calling code should NOT already have the mutex when
   # calling this.
-  def _acquire(thread)
-    if conn = available
-      @allocated[thread] = conn
+  #
+  # This should return a connection is one is available within the timeout,
+  # or nil if a connection could not be acquired within the timeout.
+  def acquire(thread)
+    if conn = assign_connection(thread)
+      return conn
     end
-  end
-  
-  if USE_WAITER
-    # Assigns a connection to the supplied thread, if one
-    # is available. The calling code should NOT already have the mutex when
-    # calling this.
-    #
-    # This should return a connection is one is available within the timeout,
-    # or nil if a connection could not be acquired within the timeout.
-    def acquire(thread)
+
+    timeout = @timeout
+    timer = Sequel.start_timer
+
+    sync do
+      @waiter.wait(@mutex, timeout)
+      if conn = next_available
+        return(@allocated[thread] = conn)
+      end
+    end
+
+    until conn = assign_connection(thread)
+      elapsed = Sequel.elapsed_seconds_since(timer)
+      raise_pool_timeout(elapsed) if elapsed > timeout
+
+      # :nocov:
+      # It's difficult to get to this point, it can only happen if there is a race condition
+      # where a connection cannot be acquired even after the thread is signalled by the condition variable
       sync do
-        if conn = _acquire(thread)
-          return conn
-        end
-
-        time = Time.now
-        @waiter.wait(@mutex, @timeout)
-
-        # Not sure why this is helpful, but calling Thread.pass after conditional
-        # variable access dramatically increases reliability when under heavy
-        # resource contention (almost eliminating timeouts), at a small cost to
-        # runtime performance.
-        Thread.pass
-
-        until conn = _acquire(thread)
-          deadline ||= time + @timeout
-          current_time = Time.now
-          raise(::Sequel::PoolTimeout, "timeout: #{@timeout}, elapsed: #{current_time - time}") if current_time > deadline
-          # :nocov:
-          # It's difficult to get to this point, it can only happen if there is a race condition
-          # where a connection cannot be acquired even after the thread is signalled by the condition
-          @waiter.wait(@mutex, deadline - current_time)
-          Thread.pass
-          # :nocov:
-        end
-
-        conn
-      end
-    end
-  else
-    # :nocov:
-    def acquire(thread)
-      unless conn = sync{_acquire(thread)}
-        time = Time.now
-        timeout = time + @timeout
-        sleep_time = @sleep_time
-        sleep sleep_time
-        until conn = sync{_acquire(thread)}
-          raise(::Sequel::PoolTimeout, "timeout: #{@timeout}, elapsed: #{Time.now - time}") if Time.now > timeout
-          sleep sleep_time
+        @waiter.wait(@mutex, timeout - elapsed)
+        if conn = next_available
+          return(@allocated[thread] = conn)
         end
       end
-      conn
+      # :nocov:
     end
-    # :nocov:
+
+    conn
   end
 
-  # Returns an available connection. If no connection is
-  # available, tries to create a new connection. The calling code should already
-  # have the mutex before calling this.
-  def available
-    next_available || make_new(DEFAULT_SERVER)
+  # Assign a connection to the thread, or return nil if one cannot be assigned.
+  # The caller should NOT have the mutex before calling this.
+  def assign_connection(thread)
+    # Thread safe as instance variable is only assigned to local variable
+    # and not operated on outside mutex.
+    allocated = @allocated
+    do_make_new = false
+    to_disconnect = nil
+
+    sync do
+      if conn = next_available
+        return(allocated[thread] = conn)
+      end
+
+      if (n = _size) >= (max = @max_size)
+        allocated.keys.each do |t|
+          unless t.alive?
+            (to_disconnect ||= []) << allocated.delete(t)
+          end
+        end
+        n = nil
+      end
+
+      if (n || _size) < max
+        do_make_new = allocated[thread] = true
+      end
+    end
+
+    if to_disconnect
+      to_disconnect.each{|dconn| disconnect_connection(dconn)}
+    end
+
+    # Connect to the database outside of the connection pool mutex,
+    # as that can take a long time and the connection pool mutex
+    # shouldn't be locked while the connection takes place.
+    if do_make_new
+      begin
+        conn = make_new(:default)
+        sync{allocated[thread] = conn}
+      ensure
+        unless conn
+          sync{allocated.delete(thread)}
+        end
+      end
+    end
+
+    conn
   end
 
   # Return a connection to the pool of available connections, returns the connection.
   # The calling code should already have the mutex before calling this.
   def checkin_connection(conn)
     @available_connections << conn
-    if USE_WAITER
-      @waiter.signal
-      Thread.pass
-    end
     conn
   end
 
-  unless method_defined?(:default_make_new)
-    # Alias the default make_new method, so subclasses can call it directly.
-    alias default_make_new make_new
-  end
-  
-  # Creates a new connection to the given server if the size of the pool for
-  # the server is less than the maximum size of the pool. The calling code
-  # should already have the mutex before calling this.
-  def make_new(server)
-    if (n = size) >= @max_size
-      @allocated.keys.each{|t| release(t) unless t.alive?}
-      n = nil
-    end
-    super if (n || size) < @max_size
-  end
-
   # Return the next available connection in the pool, or nil if there
   # is not currently an available connection.  The calling code should already
   # have the mutex before calling this.
@@ -242,9 +243,25 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
     sync{@allocated[thread]}
   end
   
-  # Create the maximum number of connections immediately.
-  def preconnect
-    (max_size - size).times{checkin_connection(make_new(nil))}
+  # Create the maximum number of connections immediately.  The calling code should
+  # NOT have the mutex before calling this.
+  def preconnect(concurrent = false)
+    enum = (max_size - _size).times
+
+    conns = if concurrent
+      enum.map{Thread.new{make_new(:default)}}.map(&:value)
+    else
+      enum.map{make_new(:default)}
+    end
+
+    sync{conns.each{|conn| checkin_connection(conn)}}
+  end
+
+  # Raise a PoolTimeout error showing the current timeout, the elapsed time, and the
+  # database's name (if any).
+  def raise_pool_timeout(elapsed)
+    name = db.opts[:name]
+    raise ::Sequel::PoolTimeout, "timeout: #{@timeout}, elapsed: #{elapsed}#{", database name: #{name}" if name}"
   end
   
   # Releases the connection assigned to the supplied thread back to the pool.
@@ -252,11 +269,12 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
   def release(thread)
     conn = @allocated.delete(thread)
 
-    if @connection_handling == :disconnect
-      db.disconnect_connection(conn)
-    else
+    unless @connection_handling == :disconnect
       checkin_connection(conn)
     end
+
+    @waiter.signal
+    nil
   end
 
   # Yield to the block while inside the mutex. The calling code should NOT
@@ -264,6 +282,4 @@ class Sequel::ThreadedConnectionPool < Sequel::ConnectionPool
   def sync
     @mutex.synchronize{yield}
   end
-  
-  CONNECTION_POOL_MAP[[false, false]] = self
 end
