@@ -13,11 +13,11 @@ module Sequel
     # The nested_attributes call defines a single method, <tt><i>association</i>_attributes=</tt>,
     # (e.g. <tt>albums_attributes=</tt>).  So if you have an Artist instance:
     #
-    #   a = Artist.new(:name=>'YJM')
+    #   a = Artist.new(name: 'YJM')
     #
     # You can create new album instances related to this artist:
     #
-    #   a.albums_attributes = [{:name=>'RF'}, {:name=>'MO'}]
+    #   a.albums_attributes = [{name: 'RF'}, {name: 'MO'}]
     #
     # Note that this doesn't send any queries to the database yet.  That doesn't happen till
     # you save the object:
@@ -33,7 +33,7 @@ module Sequel
     # objects.  You just need to make sure that the primary key field is filled in for the
     # associated object:
     #
-    #   a.update(:albums_attributes => [{:id=>1, :name=>'T'}])
+    #   a.update(:albums_attributes => [{id: 1, name: 'T'}])
     #
     # Since the primary key field is filled in, the plugin will update the album with id 1 instead
     # of creating a new album.
@@ -41,15 +41,15 @@ module Sequel
     # If you would like to delete the associated object instead of updating it, you add a _delete
     # entry to the hash, and also pass the :destroy option when calling +nested_attributes+:
     #
-    #   Artist.nested_attributes :albums, :destroy=>true
-    #   a.update(:albums_attributes => [{:id=>1, :_delete=>true}])
+    #   Artist.nested_attributes :albums, destroy: true
+    #   a.update(:albums_attributes => [{id: 1, _delete: true}])
     #
     # This will delete the related associated object from the database.  If you want to leave the
     # associated object in the database, but just remove it from the association, add a _remove
     # entry in the hash, and also pass the :remove option when calling +nested_attributes+:
     #
-    #   Artist.nested_attributes :albums, :remove=>true
-    #   a.update(:albums_attributes => [{:id=>1, :_remove=>true}])
+    #   Artist.nested_attributes :albums, remove: true
+    #   a.update(:albums_attributes => [{id: 1, _remove: true}])
     #
     # The above example was for a one_to_many association, but the plugin also works similarly
     # for other association types.  For one_to_one and many_to_one associations, you need to
@@ -66,11 +66,11 @@ module Sequel
     #
     # Your web stack will probably parse that into a nested hash similar to:
     #
-    #   {:artist=>{:name=>'Y', :albums_attributes=>{0=>{:name=>'X'}, 1=>{:id=>'2', :name=>'Z'}}}}
+    #   {'artist'=>{'name'=>'Y', 'albums_attributes'=>{'0'=>{'name'=>'X'}, '1'=>{'id'=>'2', 'name'=>'Z'}}}}
     #
     # Then you can do:
     #
-    #   artist.update(params[:artist])
+    #   artist.update_fields(params['artist'], %w'name albums_artists')
     #
     # To save changes to the artist, create the first album and associate it to the artist,
     # and update the other existing associated album.
@@ -81,10 +81,13 @@ module Sequel
       end
       
       module ClassMethods
-        # Module to store the nested_attributes setter methods, so they can
-        # call be overridden and call super to get the default behavior
-        attr_accessor :nested_attributes_module
-        
+        # Freeze nested_attributes_module when freezing model class.
+        def freeze
+          @nested_attributes_module.freeze if @nested_attributes_module
+
+          super
+        end
+
         # Allow nested attributes to be set for the given associations.  Options:
         # :destroy :: Allow destruction of nested records.
         # :fields :: If provided, should be an Array or proc. If it is an array,
@@ -110,12 +113,12 @@ module Sequel
         #
         # If a block is provided, it is used to set the :reject_if option.
         def nested_attributes(*associations, &block)
-          include(self.nested_attributes_module ||= Module.new) unless nested_attributes_module
-          opts = associations.last.is_a?(Hash) ? associations.pop : {}
+          include(@nested_attributes_module ||= Module.new) unless @nested_attributes_module
+          opts = associations.last.is_a?(Hash) ? associations.pop : OPTS
           reflections = associations.map{|a| association_reflection(a) || raise(Error, "no association named #{a} for #{self}")}
           reflections.each do |r|
-            r[:nested_attributes] = opts
-            r[:nested_attributes][:unmatched_pk] ||= opts.delete(:strict) == false ? :ignore : :raise
+            r[:nested_attributes] = opts.dup
+            r[:nested_attributes][:unmatched_pk] ||= :raise
             r[:nested_attributes][:reject_if] ||= block
             def_nested_attribute_method(r)
           end
@@ -126,7 +129,7 @@ module Sequel
         # Add a nested attribute setter method to a module included in the
         # class.
         def def_nested_attribute_method(reflection)
-          nested_attributes_module.class_eval do
+          @nested_attributes_module.class_eval do
             define_method("#{reflection[:name]}_attributes=") do |v|
               set_nested_attributes(reflection[:name], v)
             end
@@ -172,8 +175,9 @@ module Sequel
           nested_attributes_set_attributes(meta, obj, attributes)
           delay_validate_associated_object(reflection, obj)
           if reflection.returns_array?
-            send(reflection[:name]) << obj
-            after_save_hook{send(reflection.add_method, obj)}
+            public_send(reflection[:name]) << obj
+            obj.skip_validation_on_next_save!
+            after_save_hook{public_send(reflection[:add_method], obj)}
           else
             associations[reflection[:name]] = obj
 
@@ -185,9 +189,9 @@ module Sequel
             # Don't need to validate the object twice if :validate association option is not false
             # and don't want to validate it at all if it is false.
             if reflection[:type] == :many_to_one 
-              before_save_hook{send(reflection.setter_method, obj.save(:validate=>false))}
+              before_save_hook{public_send(reflection[:setter_method], obj.save(:validate=>false))}
             else
-              after_save_hook{send(reflection.setter_method, obj)}
+              after_save_hook{public_send(reflection[:setter_method], obj)}
             end
           end
           add_reciprocal_object(reflection, obj)
@@ -199,7 +203,7 @@ module Sequel
         # If there is a limit on the nested attributes for this association,
         # make sure the length of the attributes_list is not greater than the limit.
         def nested_attributes_list_setter(meta, attributes_list)
-          attributes_list = attributes_list.sort_by(&:to_s).map{|k,v| v} if attributes_list.is_a?(Hash)
+          attributes_list = attributes_list.sort.map{|k,v| v} if attributes_list.is_a?(Hash)
           if (limit = meta[:limit]) && attributes_list.length > limit
             raise(Error, "number of nested attributes (#{attributes_list.length}) exceeds the limit (#{limit})")
           end
@@ -215,9 +219,9 @@ module Sequel
           if !opts[:destroy] || reflection.remove_before_destroy?
             before_save_hook do
               if reflection.returns_array?
-                send(reflection.remove_method, obj)
+                public_send(reflection[:remove_method], obj)
               else
-                send(reflection.setter_method, nil)
+                public_send(reflection[:setter_method], nil)
               end
             end
           end
@@ -233,7 +237,7 @@ module Sequel
         def nested_attributes_set_attributes(meta, obj, attributes)
           if fields = meta[:fields]
             fields = fields.call(obj) if fields.respond_to?(:call)
-            obj.set_only(attributes, fields)
+            obj.set_fields(attributes, fields, :missing=>:skip)
           else
             obj.set(attributes)
           end
@@ -261,7 +265,7 @@ module Sequel
           str_keys = sym_keys.map(&:to_s)
           if (pk = attributes.values_at(*sym_keys)).all? || (pk = attributes.values_at(*str_keys)).all?
             pk = pk.map(&:to_s)
-            obj = Array(send(reflection[:name])).find{|x| Array(x.pk).map(&:to_s) == pk}
+            obj = Array(public_send(reflection[:name])).find{|x| Array(x.pk).map(&:to_s) == pk}
           end
           if obj
             attributes = attributes.dup.delete_if{|k,v| str_keys.include? k.to_s}
