@@ -54,7 +54,7 @@ module Sequel
             convert_output_datetime_other(v, output_timezone)
           end
         else
-          v.send(output_timezone == :utc ? :getutc : :getlocal)
+          v.public_send(output_timezone == :utc ? :getutc : :getlocal)
         end
       else
         v
@@ -69,9 +69,9 @@ module Sequel
         if v.is_a?(Date) && !v.is_a?(DateTime)
           # Dates handled specially as they are assumed to already be in the application_timezone
           if datetime_class == DateTime
-            DateTime.civil(v.year, v.month, v.day, 0, 0, 0, application_timezone == :local ? (defined?(Rational) ? Rational(Time.local(v.year, v.month, v.day).utc_offset, 86400) : Time.local(v.year, v.month, v.day).utc_offset/86400.0) : 0)
+            DateTime.civil(v.year, v.month, v.day, 0, 0, 0, application_timezone == :local ? Rational(Time.local(v.year, v.month, v.day).utc_offset, 86400) : 0)
           else
-            Time.send(application_timezone == :utc ? :utc : :local, v.year, v.month, v.day)
+            Time.public_send(application_timezone == :utc ? :utc : :local, v.year, v.month, v.day)
           end
         else
           convert_output_timestamp(convert_input_timestamp(v, input_timezone), application_timezone)
@@ -149,14 +149,14 @@ module Sequel
       when Array
         y, mo, d, h, mi, s, ns, off = v
         if datetime_class == DateTime
-          s += (defined?(Rational) ? Rational(ns, 1000000000) : ns/1000000000.0) if ns
+          s += Rational(ns, 1000000000) if ns
           if off
             DateTime.civil(y, mo, d, h, mi, s, off)
           else
             convert_input_datetime_no_offset(DateTime.civil(y, mo, d, h, mi, s), input_timezone)
           end
         else
-          Time.send(input_timezone == :utc ? :utc : :local, y, mo, d, h, mi, s, (ns ? ns / 1000.0 : 0))
+          Time.public_send(input_timezone == :utc ? :utc : :local, y, mo, d, h, mi, s, (ns ? ns / 1000.0 : 0))
         end
       when Hash
         ary = [:year, :month, :day, :hour, :minute, :second, :nanos].map{|x| (v[x] || v[x.to_s]).to_i}
@@ -164,30 +164,17 @@ module Sequel
           ary << offset
         end
         convert_input_timestamp(ary, input_timezone)
-        convert_input_timestamp(ary, input_timezone)
       when Time
         if datetime_class == DateTime
-          if v.respond_to?(:to_datetime)
-            v.to_datetime
-          else
-          # :nocov:
-            # Ruby 1.8 code, %N not available and %z broken on Windows
-            offset_hours, offset_minutes = (v.utc_offset/60).divmod(60)
-            string_to_datetime(v.strftime("%Y-%m-%dT%H:%M:%S") << sprintf(".%06i%+03i%02i", v.usec, offset_hours, offset_minutes))
-          # :nocov:
-          end
+          v.to_datetime
         else
           v
         end
       when DateTime
         if datetime_class == DateTime
           v
-        elsif v.respond_to?(:to_time)
-          v.to_time
         else
-        # :nocov:
-          string_to_datetime(v.strftime("%FT%T.%N%z"))
-        # :nocov:
+          v.to_time
         end
       else
         raise InvalidValue, "Invalid convert_input_timestamp type: #{v.inspect}"
@@ -215,7 +202,7 @@ module Sequel
     # Caches offset conversions to avoid excess Rational math.
     def time_offset_to_datetime_offset(offset_secs)
       @local_offsets ||= {}
-      @local_offsets[offset_secs] ||= respond_to?(:Rational, true) ? Rational(offset_secs, 60*60*24) : offset_secs/60/60/24.0
+      @local_offsets[offset_secs] ||= Rational(offset_secs, 86400)
     end
   end
 
