@@ -5,8 +5,8 @@ require 'version'
 Dir.chdir(File.dirname(__FILE__))
 erb = ERB.new(File.read('layout.html.erb'))
 Dir['pages/*.html.erb'].each do |page|
-  public_loc = "#{page.gsub(/\Apages\//, 'public/').gsub('.erb', '')}"
+  public_loc = "#{page.gsub(/\Apages\//, 'public/').sub('.erb', '')}"
   content = ERB.new(File.read(page)).result(binding)
-  title = File.basename(page)
+  title = File.basename(page.sub('.html.erb', ''))
   File.open(public_loc, 'wb'){|f| f.write(erb.result(binding))}
 end
